import anndata
import fsspec
import numpy as np
import pandas as pd
import pytest
import scanpy as sc
import scipy.sparse
from cirrocumulus.anndata_util import get_base
from cirrocumulus.diff_exp import DE
from cirrocumulus.parquet_dataset import ParquetDataset
from cirrocumulus.prepare_data import PrepareData
from cirrocumulus.zarr_dataset import ZarrDataset
from numpy.random import negative_binomial, binomial, seed
from scipy import sparse as sp


def get_example_data(sparse=False):
    seed(1234)
    # create test object
    adata = anndata.AnnData(
        np.multiply(binomial(1, 0.15, (100, 20)), negative_binomial(2, 0.25, (100, 20)))
    )
    # adapt marker_genes for cluster (so as to have some form of reasonable input
    adata.X[0:10, 0:5] = np.multiply(
        binomial(1, 0.9, (10, 5)), negative_binomial(1, 0.5, (10, 5))
    )

    # The following construction is inefficient, but makes sure that the same data is used in the sparse case
    if sparse:
        adata.X = sp.csr_matrix(adata.X)

    # Create cluster according to groups
    adata.obs['sc_groups'] = pd.Categorical(
        np.concatenate(
            (
                np.zeros((10,), dtype=int),
                np.ones((90,), dtype=int),
            )
        )
    )

    return adata


@pytest.fixture(autouse=True, params=[True, False])
def sparse(request):
    return request.param


def diff_results(adata, obs_field, results, group='0'):
    sc.tl.rank_genes_groups(adata, obs_field, method='t-test', pts=True)
    rank_genes_groups = adata.uns['rank_genes_groups']
    sc_scores = rank_genes_groups['scores'][group]
    sc_pvals = rank_genes_groups['pvals'][group]
    sc_pts = rank_genes_groups['pts'][group]
    sc_lfc = rank_genes_groups['logfoldchanges'][group]

    sc_df = pd.DataFrame(index=rank_genes_groups['names'][group],
                         data={'pvals': sc_pvals, 'scores': sc_scores, 'pts': sc_pts, 'lfc': sc_lfc})
    sc_df = sc_df.loc[adata.var.index]
    np.testing.assert_allclose(sc_df['pvals'], results['pvals'])

    np.testing.assert_allclose(sc_df['scores'], results['scores'], atol=1e-015)
    np.testing.assert_allclose(sc_df['lfc'], results['logfoldchanges'], atol=1e-015)

    if scipy.sparse.issparse(adata.X):
        np.testing.assert_allclose(sc_df['pts'], results['frac_expressed1'])
    else:
        assert results['frac_expressed1'] is None


@pytest.mark.parametrize("file_format", ['zarr', 'parquet'])
def test_de_backed(sparse, file_format, tmp_path):
    fs = fsspec.filesystem('file')
    adata = get_example_data(sparse)
    output_dir = str(tmp_path)
    prepare_data = PrepareData(datasets=[adata], output=output_dir, output_format=file_format)
    prepare_data.execute()
    if file_format == 'parquet':
        reader = ParquetDataset()
    elif file_format == 'zarr':
        reader = ZarrDataset()
    batch_size = 30
    obs_field = 'sc_groups'
    nfeatures = adata.shape[1]

    def get_batch_fn(i):
        end = min(nfeatures, i + batch_size)
        return reader.read_dataset(filesystem=fs, path=output_dir, dataset=dict(id=''),
                                   keys=dict(X=[slice(i, end)]))

    results = DE(series=adata.obs[obs_field], nfeatures=nfeatures, batch_size=batch_size, get_batch_fn=get_batch_fn,
                 base=get_base(adata), one_vs_rest=True)
    diff_results(adata, obs_field, results.pair2results[0])


def test_de_2_groups(sparse):
    adata = get_example_data(sparse)
    batch_size = 3
    obs_field = 'sc_groups'
    nfeatures = adata.shape[1]
    get_batch_fn = lambda i: adata[:, i:min(nfeatures, i + batch_size)]

    results = DE(series=adata.obs[obs_field], nfeatures=nfeatures, batch_size=batch_size, get_batch_fn=get_batch_fn,
                 base=get_base(adata), one_vs_rest=True)
    diff_results(adata, obs_field, results.pair2results[0])


def test_de_4_groups(sparse):
    adata1 = get_example_data(sparse)
    adata2 = get_example_data(sparse)
    adata2.obs['sc_groups'] = adata2.obs['sc_groups'].replace({0: 2, 1: 3})
    adata = anndata.concat((adata1, adata2))
    adata.obs_names_make_unique()
    batch_size = 3
    obs_field = 'sc_groups'
    adata.obs[obs_field] = adata.obs[obs_field].astype('category')
    nfeatures = adata.shape[1]
    get_batch_fn = lambda i: adata[:, i:min(nfeatures, i + batch_size)]
    de = DE(series=adata.obs[obs_field], nfeatures=nfeatures, batch_size=batch_size, get_batch_fn=get_batch_fn,
            base=get_base(adata))
    for i in range(4):
        diff_results(adata, obs_field, de.pair2results[i], str(i))



import logging
import math
import os

import pandas as pd
import pandas._libs.json as ujson

from cirrocumulus.diff_exp import DE
from .data_processing import get_filter_str, get_mask
from .envir import CIRRO_SERVE, CIRRO_MAX_WORKERS, CIRRO_DATABASE_CLASS, CIRRO_JOB_RESULTS, CIRRO_JOB_TYPE
from .fdr import fdrcorrection
from .util import create_instance, add_dataset_providers, get_fs, import_path, open_file

executor = None
job_id_2_future = dict()

logger = logging.getLogger('cirro')


def save_job_result_to_file(result, job_id):
    new_result = dict()
    new_result['content-type'] = result.pop('content-type')
    if new_result['content-type'] == 'application/json':
        new_result['content-encoding'] = 'gzip'
        url = os.path.join(os.environ[CIRRO_JOB_RESULTS], str(job_id) + '.json.gz')
        with open_file(url, 'wt', compression='gzip') as out:
            out.write(ujson.dumps(result, double_precision=2, orient='values'))
    elif new_result['content-type'] == 'application/h5ad':
        url = os.path.join(os.environ[CIRRO_JOB_RESULTS], str(job_id) + '.h5ad')
        with get_fs(url).open(url, 'wb') as out:
            result['data'].write(out)
    elif new_result['content-type'] == 'application/zarr':
        url = os.path.join(os.environ[CIRRO_JOB_RESULTS], str(job_id) + '.zarr')
        result['data'].write_zarr(get_fs(url).get_mapper(url))
    elif new_result['content-type'] == 'application/parquet':
        import pyarrow.parquet as pq
        import pyarrow as pa
        url = os.path.join(os.environ[CIRRO_JOB_RESULTS], str(job_id) + '.parquet')
        pq.write_table(pa.Table.from_pandas(result['data']), url, filesystem=get_fs(url))
    else:
        raise ValueError('Unknown content-type {}'.format(new_result['content-type']))
    new_result['url'] = url
    return new_result


def delete_job(job_id):
    future = job_id_2_future.get(job_id)
    if future is not None and not future.done():
        del job_id_2_future[job_id]
        future.cancel()
        logger.info('Cancel job {}'.format(job_id))


def done_callback(future):
    for job_id in list(job_id_2_future.keys()):
        if job_id_2_future[job_id] == future:
            del job_id_2_future[job_id]
            logger.info('Job {} done'.format(job_id))
            break


def submit_job(database_api, dataset_api, email, dataset, job_name, job_type, params):
    global executor

    is_serve = os.environ.get(CIRRO_SERVE) == 'true'
    if executor is None:
        max_workers = int(os.environ.get(CIRRO_MAX_WORKERS, '2' if is_serve else '1'))
        if max_workers > 0:
            import multiprocessing
            from concurrent.futures.process import ProcessPoolExecutor
            from concurrent.futures.thread import ThreadPoolExecutor
            executor = ProcessPoolExecutor(max_workers=max_workers, mp_context=multiprocessing.get_context(
                'spawn')) if is_serve else ThreadPoolExecutor(
                max_workers=max_workers)
    job_id = database_api.create_job(email=email, dataset_id=dataset['id'], job_name=job_name, job_type=job_type,
                                     params=params)
    if executor is not None:
        future = executor.submit(run_job, email, job_id, job_type, dataset, params,
                                 database_api if not is_serve else None,
                                 dataset_api if not is_serve else None)
        future.add_done_callback(done_callback)
        job_id_2_future[job_id] = future
    else:
        run_job(email, job_id, job_type, dataset, params, database_api if not is_serve else None,
                dataset_api if not is_serve else None)
    return job_id


def get_obs(dataset_api, dataset, dataset_info, params):
    obs_fields = params.get('obs')
    if obs_fields is not None:
        obs = dataset_api.read_dataset(keys=dict(obs=obs_fields), dataset=dataset).obs
        obs_field = obs_fields[0]
        if len(obs_fields) > 1:
            # combine in to one field
            obs_field = '_'.join(obs_fields)
            obs[obs_field] = obs[obs_fields[0]].astype(str)
            for i in range(1, len(obs_fields)):
                obs[obs_field] += '_' + obs[obs_fields[i]].astype(str)
            obs[obs_field] = obs[obs_field].astype('category')
        return obs, obs_field
    else:
        filters = [params['filter'], params['filter2']]
        filter_names = [get_filter_str(params['filter']), get_filter_str(params['filter2'])]
        for i in range(len(filter_names)):
            if filter_names[i] is None:
                filter_names[i] = 'group_' + str(i + 1)
        obs = pd.DataFrame(index=pd.RangeIndex(dataset_info['shape'][0]).astype(str))
        obs_field = 'selection'
        obs[obs_field] = '3'
        masks, _ = get_mask(dataset_api, dataset, filters)
        for i in range(len(masks)):
            obs.loc[masks[i], obs_field] = filter_names[i]
        obs[obs_field] = obs[obs_field].astype('category')
        return obs, obs_field


def run_job(email, job_id, job_type, dataset, params, database_api, dataset_api):
    if database_api is None:
        database_api = create_instance(os.environ[CIRRO_DATABASE_CLASS])
    if dataset_api is None:
        from cirrocumulus.api import dataset_api
        add_dataset_providers()
    database_api.update_job(email=email, job_id=job_id, status='running', result=None)
    f = os.environ[CIRRO_JOB_TYPE + job_type]
    if f is None:
        database_api.update_job(email=email, job_id=job_id, status='error', result=None)
        raise ValueError('No function to handle {}'.format(job_type))
    import_path(f)(email, job_id, job_type, dataset, params, database_api, dataset_api)


def run_de(email, job_id, job_type, dataset, params, database_api, dataset_api):
    dataset_info = dataset_api.get_dataset_info(dataset)
    var_names = dataset_info['var']
    nfeatures = len(var_names)
    batch_size = math.ceil(nfeatures / 5) if os.environ.get(
        CIRRO_SERVE) == 'true' else nfeatures  # TODO more intelligent batches

    def get_batch_fn(i):
        end = min(nfeatures, i + batch_size)
        adata = dataset_api.read_dataset(keys=dict(X=[slice(i, end)]), dataset=dataset)
        if batch_size != nfeatures:
            frac = end / nfeatures
            status = 'running {:.0f}%'.format(100 * frac) if frac < 1 else 'saving results'
            logger.info(status)
            database_api.update_job(email=email, job_id=job_id, status=status, result=None)
        return adata

    compare_pairs = params.get('pairs') == '1'
    obs, obs_field = get_obs(dataset_api, dataset, dataset_info, params)

    de_results = DE(series=obs[obs_field], nfeatures=nfeatures, batch_size=batch_size, get_batch_fn=get_batch_fn,
                    one_vs_rest=not compare_pairs)  # TODO get base

    # group:field is object entry
    result_df = pd.DataFrame(data={'index': var_names})
    has_frac_expressed = False
    comparison_names = []
    for comparison in de_results.pair2results.keys():
        result = de_results.pair2results[comparison]
        comparison_name = comparison if isinstance(comparison, str) else '_'.join(comparison)
        comparison_names.append(comparison_name)

        pvals = fdrcorrection(result['pvals'])
        result_df[f'{comparison_name}:pvals_adj'] = pvals
        result_df[f'{comparison_name}:scores'] = result['scores']
        result_df[f'{comparison_name}:lfc'] = result['logfoldchanges']

        if result.get('frac_expressed1') is not None:
            has_frac_expressed = True
            result_df[f'{comparison_name}:pts_1'] = result['frac_expressed1']
            result_df[f'{comparison_name}:pts_2'] = result['frac_expressed2']
    # client expects field {comparison_name}:pvals_adj
    result = dict(groups=comparison_names,
                  fields=['pvals_adj', 'scores', 'lfc'] + (
                      ['pts_1', 'pts_2'] if has_frac_expressed else []),
                  data=result_df.to_dict(orient='records'))
    result['content-type'] = 'application/json'
    database_api.update_job(email=email, job_id=job_id, status='complete', result=result)



CIRRO_TEST = 'CIRRO_TEST'  # whether running using pytest
CIRRO_AUTH_CLIENT_ID = 'CIRRO_AUTH_CLIENT_ID'  # for javascript client
CIRRO_AUTH_BASE_URL = 'CIRRO_AUTH_BASE_URL'  # for javascript client
CIRRO_DB_URI = 'CIRRO_DB_URI'
CIRRO_EMAIL = 'CIRRO_EMAIL'
CIRRO_SERVE = 'CIRRO_SERVE'
CIRRO_MAX_WORKERS = 'CIRRO_MAX_WORKERS'
CIRRO_FOOTER = 'CIRRO_FOOTER'
CIRRO_BRAND = 'CIRRO_BRAND'
CIRRO_UPLOAD = 'CIRRO_UPLOAD'
CIRRO_JOB_RESULTS = 'CIRRO_JOB_RESULTS'
CIRRO_DATABASE_CLASS = 'CIRRO_DATABASE_CLASS'
CIRRO_DATABASE = 'CIRRO_DATABASE'
CIRRO_DATASET_PROVIDERS = 'CIRRO_DATASET_PROVIDERS'
CIRRO_AUTH = 'CIRRO_AUTH'  # provider to validate tokens
CIRRO_MIXPANEL = 'CIRRO_MIXPANEL'
# for mounting a bucket locally. Comma separated string of bucket:local_path. Example s3://foo/bar:/fsx
CIRRO_MOUNT = 'CIRRO_MOUNT'
CIRRO_LOG_LEVEL = 'CIRRO_LOG_LEVEL'
# columns to display to user
CIRRO_DATASET_SELECTOR_COLUMNS = 'CIRRO_DATASET_SELECTOR_COLUMNS'
CIRRO_JOB_TYPE = 'CIRRO_JOB_TYPE'

# path to JSON file for library list when adding new dataset
CIRRO_LIBRARY = 'CIRRO_LIBRARY'

# path to JSON file for species list when adding new dataset
CIRRO_SPECIES = 'CIRRO_SPECIES'
# path to OBO file
CIRRO_CELL_ONTOLOGY = 'CIRRO_CELL_ONTOLOGY'

# comma separated list of paths to allow all logged in users to download from
CIRRO_STATIC_DIR = 'CIRRO_STATIC_DIR'

SERVER_CAPABILITY_RENAME_CATEGORIES = 'SERVER_CAPABILITY_RENAME_CATEGORIES'
SERVER_CAPABILITY_JOBS = 'SERVER_CAPABILITY_JOBS'
SERVER_CAPABILITY_FEATURE_SETS = 'SERVER_CAPABILITY_FEATURE_SETS'
SERVER_CAPABILITY_LINKS = 'SERVER_CAPABILITY_LINKS'
SERVER_CAPABILITY_EDIT_DATASET = 'SERVER_CAPABILITY_EDIT_DATASET'
SERVER_CAPABILITY_ADD_DATASET = 'SERVER_CAPABILITY_ADD_DATASET'
SERVER_CAPABILITY_DELETE_DATASET = 'SERVER_CAPABILITY_DELETE_DATASET'



import datetime
import json
import os

from cirrocumulus.abstract_db import AbstractDB
from cirrocumulus.envir import SERVER_CAPABILITY_JOBS, CIRRO_EMAIL, CIRRO_JOB_RESULTS
from cirrocumulus.invalid_usage import InvalidUsage
from cirrocumulus.job_api import save_job_result_to_file
from cirrocumulus.util import get_email_domain
from google.cloud import datastore

DATASET = 'Dataset'
CAT_NAME = 'Cat_Name'
DATASET_FILTER = 'Dataset_Filter'
DATASET_VIEW = 'D_View'
FEATURE_SET = 'F_Set'
USER = 'User'
JOB = 'Job'


def get_datasets(results, email, query, unique_ids):
    for result in query.fetch():
        if result.id not in unique_ids:
            unique_ids.add(result.id)
            results.append(
                {'id': result.id, 'name': result['name'], 'title': result.get('title'),
                 'owner': email in result['owners'], 'species': result.get('species'), 'url': result['url']})


class CloudFireStoreNative(AbstractDB):

    def __init__(self):
        super().__init__()
        self.datastore_client = datastore.Client()
        os.environ[CIRRO_EMAIL] = self.datastore_client.project + '@appspot.gserviceaccount.com'

    def capabilities(self):
        c = super().capabilities()
        c[SERVER_CAPABILITY_JOBS] = False
        return c

    def __get_key_and_dataset(self, email, dataset_id, ensure_owner=False):
        client = self.datastore_client
        key = client.key(DATASET, int(dataset_id))
        dataset = client.get(key)
        if dataset is None:
            raise InvalidUsage('Please provide a valid id', 400)
        readers = dataset.get('readers')

        domain = get_email_domain(email)
        if email not in readers and domain not in readers:
            raise InvalidUsage('Not authorized', 403)
        if ensure_owner and email not in dataset['owners']:
            raise InvalidUsage('Not authorized', 403)
        return key, dataset

    def __get_entity_list(self, email, dataset_id, kind, keys):
        client = self.datastore_client
        self.__get_key_and_dataset(email, dataset_id)
        query = client.query(kind=kind)
        query.add_filter('dataset_id', '=', int(dataset_id))
        results = []
        for result in query.fetch():
            d = dict(id=result.id)
            for key in keys:
                d[key] = result.get(key)
            results.append(result)
        return results

    def __delete_entity(self, email, kind, entity_id):
        client = self.datastore_client
        key = client.key(kind, int(entity_id))
        e = client.get(key)
        self.__get_key_and_dataset(email, e['dataset_id'])
        client.delete(key)

    def __get_entity(self, email, kind, entity_id):
        client = self.datastore_client
        e = client.get(client.key(kind, int(entity_id)))
        self.__get_key_and_dataset(email, e['dataset_id'])
        return e

    def __upsert_entity(self, email, dataset_id, entity_id, kind, entity_update):
        dataset_id = int(dataset_id)
        client = self.datastore_client
        self.__get_key_and_dataset(email, dataset_id)
        if entity_id is None:
            e = datastore.Entity(client.key(kind),
                                 exclude_from_indexes=['value'])
        else:
            entity_id = int(entity_id)
            key = client.key(kind, entity_id)
            e = client.get(key)

        if email is not None:
            entity_update['email'] = email
        if dataset_id is not None:
            entity_update['dataset_id'] = dataset_id

        e.update(entity_update)
        client.put(e)
        return e.id

    def category_names(self, email, dataset_id):
        dataset_id = int(dataset_id)
        self.__get_key_and_dataset(email, dataset_id, False)
        client = self.datastore_client
        query = client.query(kind=CAT_NAME)
        query.add_filter('dataset_id', '=', dataset_id)
        results = {}
        for result in query.fetch():
            category = result['category']
            cat_results = results.get(category)
            if cat_results is None:
                cat_results = {}
                results[category] = cat_results
            cat_results[result['original']] = {'newValue': result.get('new'),
                                               'negativeMarkers': result.get('negativeMarkers'),
                                               'positiveMarkers': result.get('positiveMarkers'),
                                               'color': result.get('color')}

        return results

    def upsert_category_name(self, email, dataset_id, category, original_value, update):
        dataset_id = int(dataset_id)
        self.__get_key_and_dataset(email, dataset_id, False)
        client = self.datastore_client
        key = client.key(CAT_NAME, str(dataset_id) + '-' + str(category) + '-' + str(original_value))
        entity = datastore.Entity(key=key)
        update['category'] = category
        update['dataset_id'] = dataset_id
        update['original'] = original_value
        if 'newValue' in update:  # backwards compatibility
            update['new'] = update.pop('newValue')
        entity.update(update)
        client.put(entity)
        return entity.id

    def user(self, email):
        client = self.datastore_client
        key = client.key(USER, email)
        user = client.get(key)
        if user is None:
            user = datastore.Entity(client.key(USER, email))
        user.update({'last_login': datetime.datetime.utcnow()})
        client.put(user)
        return user

    def datasets(self, email):
        client = self.datastore_client
        query = client.query(kind=DATASET)
        query.add_filter('readers', '=', email)
        results = []
        unique_ids = set()
        get_datasets(results, email, query, unique_ids)
        domain = get_email_domain(email)
        if domain is not None:
            query = client.query(kind=DATASET)
            query.add_filter('readers', '=', domain)
            # TODO make this more efficient
            get_datasets(results, email, query, unique_ids)

            query = client.query(kind=DATASET)
            query.add_filter('readers', '=', 'allUsers')
            get_datasets(results, email, query, unique_ids)
        return results

    def delete_dataset(self, email, dataset_id):
        client = self.datastore_client
        key, dataset = self.__get_key_and_dataset(email, dataset_id, True)
        client.delete(key)

    def get_dataset(self, email, dataset_id, ensure_owner=False):
        _, dataset = self.__get_key_and_dataset(email, dataset_id, ensure_owner)
        dataset['id'] = dataset.id
        return dataset

    def is_importer(self, email):
        # TODO check if user can modify dataset
        client = self.datastore_client
        user = client.get(client.key(USER, email))
        if 'importer' not in user:
            raise False
        return user['importer']

    def upsert_dataset(self, email, readers, dataset):
        client = self.datastore_client

        if readers is not None:
            readers = set(readers)
            if email in readers:
                readers.remove(email)
            readers.add(email)
            dataset['readers'] = list(readers)
        dataset_id = dataset.get('id')
        if dataset_id is not None:  # only owner can update
            key, dataset_entity = self.__get_key_and_dataset(email, dataset_id, True)
        else:  # new dataset
            dataset_entity = datastore.Entity(client.key(DATASET), exclude_from_indexes=['url'])
            user = client.get(client.key(USER, email))
            if 'importer' not in user or not user['importer']:
                raise InvalidUsage('Not authorized', 403)
            dataset['owners'] = [email]

        dataset_entity.update(dataset)
        client.put(dataset_entity)
        dataset_id = dataset.id
        return dataset_id

    # feature sets
    def get_feature_sets(self, email, dataset_id):
        return self.__get_entity_list(email=email, dataset_id=dataset_id, kind=FEATURE_SET,
                                      keys=['category', 'name', 'features'])

    def delete_feature_set(self, email, dataset_id, set_id):
        return self.__delete_entity(email=email, kind=FEATURE_SET, entity_id=set_id)

    def upsert_feature_set(self, email, dataset_id, set_id, category, name, features):
        entity_update = {}
        if name is not None:
            entity_update['name'] = name
        if category is not None:
            entity_update['category'] = category
        if features is not None:
            entity_update['features'] = features
        return self.__upsert_entity(email=email, dataset_id=dataset_id, entity_id=set_id, kind=FEATURE_SET,
                                    entity_update=entity_update)

    # views
    def dataset_views(self, email, dataset_id):
        return self.__get_entity_list(email=email, dataset_id=dataset_id, kind=DATASET_VIEW,
                                      keys=['name', 'notes', 'email', 'last_updated'])

    def delete_dataset_view(self, email, view_id):
        return self.__delete_entity(email=email, kind=DATASET_VIEW, entity_id=view_id)

    def get_dataset_view(self, email, view_id):
        return self.__get_entity(email=email, entity_id=view_id, kind=DATASET_VIEW)

    def upsert_dataset_view(self, email, dataset_id, view):
        view['last_updated'] = datetime.datetime.utcnow()
        if 'value' in view:
            view['value'] = json.dumps(view['value'])
        if email is not None:
            view['email'] = email
        if dataset_id is not None:
            view['dataset_id'] = dataset_id
        view_id = view.pop('id') if 'id' in view else None
        view_id = self.__upsert_entity(email=email, dataset_id=dataset_id, entity_id=view_id, kind=DATASET_VIEW,
                                       entity_update=view)
        return dict(id=view_id, last_updated=view['last_updated'])

    def create_job(self, email, dataset_id, job_name, job_type, params):
        dataset_id = int(dataset_id)
        client = self.datastore_client
        self.__get_key_and_dataset(email, dataset_id)
        entity = datastore.Entity(client.key(JOB), exclude_from_indexes=["params"])
        import json
        entity.update(dict(dataset_id=dataset_id, email=email, name=job_name, type=job_type, status='',
                           submitted=datetime.datetime.utcnow(), params=json.dumps(params)))
        client.put(entity)
        job_id = entity.id
        return str(job_id)

    def get_job(self, email, job_id, return_type):
        client = self.datastore_client
        job_id = int(job_id)
        key = client.key(JOB, job_id)
        result = client.get(key)
        self.__get_key_and_dataset(email, result['dataset_id'])
        if return_type == 'result':
            return result['result']
        elif return_type == 'status':
            return dict(status=result['status'])
        elif return_type == 'params':
            return dict(params=result.get('params'))

    def update_job(self, email, job_id, status, result):
        if not self.capabilities()[SERVER_CAPABILITY_JOBS]:
            return
        client = self.datastore_client
        job_id = int(job_id)
        is_complete = result is not None
        key = client.key(JOB, job_id)
        entity = client.get(key)
        if not is_complete:
            self.__get_key_and_dataset(email, entity['dataset_id'])

        if is_complete:
            if os.environ.get(CIRRO_JOB_RESULTS) is not None:  # save to directory
                result = save_job_result_to_file(result, job_id)
            else:
                raise ValueError('Please specify job result location')
            entity.update(dict(result=result, status=status))
            client.put(entity)
        else:
            entity.update(dict(status=status))
            client.put(entity)

    def get_jobs(self, email, dataset_id):
        client = self.datastore_client
        results = []

        query = client.query(kind=JOB)
        dataset_id = int(dataset_id)
        self.__get_key_and_dataset(email, dataset_id, False)
        query.add_filter('dataset_id', '=', dataset_id)

        for result in query.fetch():
            results.append(
                dict(id=result.id, name=result['name'], type=result['type'], submitted=result.get('submitted'),
                     status=result.get('status'), email=result.get('email')))
        return results

    def delete_job(self, email, job_id):
        client = self.datastore_client
        job_id = int(job_id)
        key = client.key(JOB, job_id)
        result = client.get(key)
        if result['email'] == email:
            client.delete(key)
        else:
            raise InvalidUsage('Not authorized', 403)



import os

import fsspec
import numpy as np
import pandas as pd
import pytest
import scipy.sparse

from cirrocumulus.parquet_dataset import ParquetDataset
from cirrocumulus.prepare_data import PrepareData
from cirrocumulus.zarr_dataset import ZarrDataset


def read_and_diff(ds_reader, path, test_data, measures, dimensions, continuous_obs, basis):
    dataset = dict(id='')
    fs = fsspec.filesystem('file')
    prepared_adata = ds_reader.read_dataset(filesystem=fs, path=path, dataset=dataset,
                                            keys=dict(X=measures, obs=dimensions + continuous_obs,
                                                      basis=[basis]))

    assert scipy.sparse.issparse(test_data.X) == scipy.sparse.issparse(prepared_adata.X)
    if scipy.sparse.issparse(test_data.X):
        test_data.X = test_data.X.toarray()
        prepared_adata.X = prepared_adata.X.toarray()
    np.testing.assert_equal(test_data.X, prepared_adata.X)
    for key in dimensions + continuous_obs:
        pd.testing.assert_series_equal(test_data.obs[key], prepared_adata.obs[key], check_index=False,
                                       check_flags=False)

    np.testing.assert_equal(prepared_adata.obsm[basis], test_data.obsm[basis])
    # ensure shape is correct when reading with no keys
    prepared_adata2 = ds_reader.read_dataset(filesystem=fs, path=path, dataset=dataset, keys=dict())
    assert prepared_adata2.shape[0] == test_data.shape[0]


def test_prepare_cxg_tile_db(test_data, measures, dimensions, continuous_obs, basis, tmp_path):
    try:
        from cirrocumulus.tiledb_dataset import TileDBDataset
        output_dir = str(tmp_path)
        test_data = test_data[:, measures]
        test_data.obs = test_data.obs[dimensions + continuous_obs]
        import subprocess
        output_cxg = os.path.join(output_dir, 'test.cxg')
        output_h5ad = os.path.join(output_dir, 'test.h5ad')
        test_data.write(output_h5ad)
        subprocess.check_call(['cellxgene', 'convert', '-o', output_cxg, '--disable-corpora-schema', output_h5ad])
        read_and_diff(TileDBDataset(), output_cxg, test_data, measures, dimensions, continuous_obs, basis)
    except ModuleNotFoundError:
        print('Skipping TileDB tests')


def test_prepare_join_obs_index(test_data, tmp_path):
    output_dir = str(tmp_path)
    test_data2 = test_data.copy()
    test_data2 = test_data2[[0, 1, 2]]
    with pytest.raises(ValueError):
        PrepareData(datasets=[test_data, test_data2], output=output_dir)


@pytest.mark.parametrize("file_format", ['zarr', 'parquet'])
def test_prepare(test_data, measures, dimensions, continuous_obs, basis, file_format, tmp_path):
    file_format2ext = dict(parquet='.cpq', zarr='.zarr')
    output_dir = str(tmp_path / 'test.{}'.format(file_format2ext[file_format]))
    test_data = test_data[:, measures]
    test_data.obs = test_data.obs[dimensions + continuous_obs]
    prepare_data = PrepareData(datasets=[test_data], output=output_dir, output_format=file_format)
    prepare_data.execute()
    if file_format == 'parquet':
        reader = ParquetDataset()
    elif file_format == 'zarr':
        reader = ZarrDataset()
    read_and_diff(reader, output_dir, test_data, measures,
                  dimensions, continuous_obs, basis)


def test_prepare_jsonl(test_data, measures, dimensions, continuous_obs, basis, tmp_path):
    output_dir = str(tmp_path)
    test_data = test_data[:, measures]
    test_data.obs = test_data.obs[dimensions + continuous_obs]
    prepare_data = PrepareData(datasets=[test_data],
                               output=os.path.join(output_dir, 'test.jsonl'), output_format='jsonl')
    prepare_data.execute()



import numpy as np
import pandas as pd
import scipy.sparse

from cirrocumulus.data_processing import handle_data
from cirrocumulus.embedding_aggregator import EmbeddingAggregator


def create_df(test_data, measures, dimensions, basis):
    df = pd.DataFrame(test_data[:, measures].X.toarray(), columns=measures)
    df = df.join(test_data.obs[dimensions].reset_index())
    return df.join(pd.DataFrame(test_data.obsm['X_umap'][:, 0:2], columns=basis['coordinate_columns']))


def group_df(test_data, measures, dimensions, continuous_obs, basis):
    df = create_df(test_data, measures, dimensions + continuous_obs, basis)
    agg_dict = {}
    for key in basis['coordinate_columns']:
        agg_dict[key] = 'max'
    for key in measures + continuous_obs:
        agg_dict[key] = basis['agg']
    for key in dimensions:
        agg_dict[key] = lambda x: x.mode()[0]
    EmbeddingAggregator.convert_coords_to_bin(df, nbins=100,
                                              coordinate_columns=basis['coordinate_columns'],
                                              bin_name=basis['full_name'])
    # group by bin then agg
    return df.groupby(basis['full_name']).agg(agg_dict)


def diff_binning(grouped_df, measures, dimensions, continuous_obs, basis, results):
    for key in basis['coordinate_columns']:
        np.testing.assert_array_equal(results['coordinates'][key], grouped_df[key].values, err_msg=key)
    for key in measures:
        v = results['values'][key]  # dict of indices, values
        if isinstance(v, dict):
            values = np.zeros(len(grouped_df))
            values[v['indices']] = v['values']
        else:
            values = v
        np.testing.assert_allclose(values, grouped_df[key].values, err_msg=key, rtol=1e-05, atol=1e-06)
    for key in continuous_obs:
        np.testing.assert_allclose(results['values'][key], grouped_df[key].values, err_msg=key, rtol=1e-05,
                                   atol=1e-08)
    for key in dimensions:
        np.testing.assert_array_equal(results['values'][key]['value'], grouped_df[key].values, err_msg=key)


def test_no_binning(dataset_api, input_dataset, test_data, measures, dimensions, continuous_obs, basis):
    obsm_field = basis
    embedding_list = [dict(dimensions=2, name=basis)]
    values = dict(dimensions=dimensions, measures=measures + list(map(lambda x: 'obs/' + x, continuous_obs)))
    results = handle_data(dataset_api=dataset_api, dataset=input_dataset, values=values, embedding_list=embedding_list)
    for key in measures:
        v = results['values'][key]  # dict of indices, values
        if isinstance(v, dict):
            values = np.zeros(test_data.shape[0])
            values[v['indices']] = v['values']
        else:
            values = v
        X = test_data[:, key].X
        if scipy.sparse.issparse(X):
            X = X.toarray()
        X = X.flatten()
        np.testing.assert_array_equal(values, X, err_msg=key)
    for key in dimensions:
        val = results['values'][key]
        if pd.api.types.is_categorical_dtype(test_data.obs[key]):
            val = pd.Categorical.from_codes(val['values'], val['categories'])
        np.testing.assert_array_equal(val, test_data.obs[key].values, err_msg="obs field {}".format(key))
    obsm = test_data.obsm[obsm_field]
    coords = results['embeddings'][0]['coordinates']
    for i in range(obsm.shape[1]):
        np.testing.assert_array_equal(coords['{}_{}'.format(basis, i + 1)], obsm[:, i],
                                      err_msg="obsm {}".format(key))

#
# @pytest.fixture(autouse=True, params=['sum', 'mean', 'max'])
# def agg_function(request):
#     return request.param
#
#
# def test_binning(dataset_api, input_dataset, test_data, measures, dimensions, continuous_obs, basis,
#                  agg_function):
#     nbins = 100
#     basis_dict = get_basis(basis, nbins, agg_function)
#     grouped_df = group_df(test_data, measures, dimensions, continuous_obs, basis_dict)
#     embedding_list = [dict(nbins=nbins, ndim=2, basis=basis, dimensions=dimensions, agg=agg_function)]
#     embedding_list[0]['measures'] = measures + list(map(lambda x: 'obs/' + x, continuous_obs))
#     results = handle_data(dataset_api=dataset_api, dataset=input_dataset, embedding_list=embedding_list)
#     diff_binning(grouped_df, measures, dimensions, continuous_obs, basis_dict, results['embeddings'][0])



import os

import anndata
import pytest

from cirrocumulus.envir import CIRRO_DB_URI, CIRRO_TEST
from cirrocumulus.launch import create_app, configure_app
from cirrocumulus.prepare_data import PrepareData
from cirrocumulus.serve import cached_app


@pytest.fixture(scope='session', params=[True, False])
def app_conf(request, tmpdir_factory):
    dataset_path = 'test-data/pbmc3k_no_raw.h5ad'

    if not request.param:
        app = create_app()
        configure_app(app, [dataset_path], None, None)
        dataset_id = dataset_path
        os.environ[CIRRO_TEST] = 'false'
    else:
        os.environ[CIRRO_TEST] = 'true'
        os.environ[CIRRO_DB_URI] = 'mongodb://localhost:27018/cirrocumulus-test'
        app = cached_app()
    with app.test_client() as client:
        if request.param:
            # insert dataset
            output_dir = str(tmpdir_factory.mktemp("data").join("test.zarr"))
            PrepareData(datasets=[anndata.read(dataset_path)], output=output_dir, output_format='zarr',
                        no_auto_groups=True).execute()
            r = client.post('/api/dataset', data=dict(url=output_dir, name='test'))
            dataset_id = r.json['id']
        yield client, dataset_id


def test_server(app_conf):
    client, dataset_id = app_conf
    r = client.get('/api/server').json
    assert r['email'] is None


def test_schema(app_conf):
    client, dataset_id = app_conf
    r = client.get('/api/schema?id={}'.format(dataset_id)).json
    assert isinstance(r['var'], list)
    assert isinstance(r['obs'], list)
    assert isinstance(r['embeddings'], list)
    assert len(r['obsCat']) == 1 and r['obsCat'][0] == 'louvain'
    assert r['shape'][0] == 2638 and r['shape'][1] == 1838



import datetime
import os

import pandas._libs.json as ujson
from bson import ObjectId
from pymongo import MongoClient

from cirrocumulus.abstract_db import AbstractDB
from cirrocumulus.util import get_email_domain, get_fs
from .envir import CIRRO_DB_URI, CIRRO_AUTH_CLIENT_ID, CIRRO_JOB_RESULTS, SERVER_CAPABILITY_EDIT_DATASET, \
    SERVER_CAPABILITY_ADD_DATASET, SERVER_CAPABILITY_DELETE_DATASET, SERVER_CAPABILITY_LINKS, SERVER_CAPABILITY_JOBS, \
    SERVER_CAPABILITY_FEATURE_SETS, SERVER_CAPABILITY_RENAME_CATEGORIES, CIRRO_TEST
from .invalid_usage import InvalidUsage
from .job_api import save_job_result_to_file


def format_doc(d):
    d['id'] = str(d.pop('_id'))
    return d


class MongoDb(AbstractDB):

    def __init__(self):
        super().__init__()
        if bool(os.environ.get(CIRRO_TEST)):
            import mongomock
            self.client = mongomock.MongoClient(os.environ[CIRRO_DB_URI])
        else:
            self.client = MongoClient(os.environ[CIRRO_DB_URI])
        self.db = self.client.get_default_database()
        self.db.categories.create_index('dataset_id')
        self.db.views.create_index('dataset_id')
        self.db.feature_sets.create_index('dataset_id')
        self.db.jobs.create_index('dataset_id')
        self.db.datasets.create_index('readers')
        self.fs = None

    def category_names(self, email, dataset_id):

        self.get_dataset(email, dataset_id)
        collection = self.db.categories
        results = {}
        for doc in collection.find(dict(dataset_id=dataset_id)):
            category = doc['category']
            cat_results = results.get(category)
            if cat_results is None:
                cat_results = {}
                results[category] = cat_results
            cat_results[doc['original']] = {'newValue': doc.get('new'),
                                            'negativeMarkers': doc.get('negativeMarkers'),
                                            'positiveMarkers': doc.get('positiveMarkers'),
                                            'color': doc.get('color')}
        return results

    def upsert_category_name(self, email, dataset_id, category, original_value, update):
        if not self.capabilities()[SERVER_CAPABILITY_RENAME_CATEGORIES]:
            return
        self.get_dataset(email, dataset_id)
        collection = self.db.categories
        key = str(dataset_id) + '-' + str(category) + '-' + str(original_value)
        update['category'] = category
        update['dataset_id'] = dataset_id
        update['original'] = original_value
        update['last_updated'] = datetime.datetime.utcnow()
        if 'newValue' in update:  # backwards compatibility
            update['new'] = update.pop('newValue')
        collection.update_one(dict(cat_id=key), {'$set': update}, upsert=True)

    def user(self, email):
        collection = self.db.users
        doc = collection.find_one(dict(email=email))
        if doc is None:
            collection.insert_one(dict(email=email))
            return {'id': email}
        else:
            return {'id': email,
                    'importer': doc.get('importer', False)}

    def get_dataset(self, email, dataset_id, ensure_owner=False):
        collection = self.db.datasets
        auth_client_id = os.environ.get(CIRRO_AUTH_CLIENT_ID)

        if auth_client_id is None:  # allow unregistered URL
            try:
                dataset_id.index('://')
                return {
                    'id': dataset_id,
                    'name': dataset_id,
                    'url': dataset_id
                }
            except ValueError:
                pass
        doc = collection.find_one(dict(_id=ObjectId(dataset_id)))
        if doc is None:
            raise InvalidUsage('Please provide a valid id', 400)
        readers = doc.get('readers')
        domain = get_email_domain(email)
        if email not in readers and domain not in readers and '' not in readers:
            raise InvalidUsage('Not authorized', 403)
        if ensure_owner and email not in doc['owners']:
            raise InvalidUsage('Not authorized', 403)
        doc['owner'] = 'owners' in doc and email in doc.pop('owners')
        doc['id'] = str(doc.pop('_id'))
        return doc

    def datasets(self, email):
        collection = self.db.datasets
        results = []
        domain = get_email_domain(email)
        if domain is None:
            query = dict(readers={'$in': [email, '']})
        else:
            query = dict(readers={'$in': [email, domain]})

        for doc in collection.find(query):
            doc['owner'] = 'owners' in doc and email in doc.pop('owners')
            doc['id'] = str(doc.pop('_id'))
            results.append(doc)
        return results

    # views
    def dataset_views(self, email, dataset_id):
        self.get_dataset(email, dataset_id)
        collection = self.db.views
        results = []
        for doc in collection.find(dict(dataset_id=dataset_id), dict(value=0)):
            results.append(format_doc(doc))
        return results

    def delete_dataset_view(self, email, view_id):
        if not self.capabilities()[SERVER_CAPABILITY_LINKS]:
            return
        collection = self.db.views
        doc = collection.find_one(dict(_id=ObjectId(view_id)))
        self.get_dataset(email, doc['dataset_id'])
        collection.delete_one(dict(_id=ObjectId(view_id)))

    def get_dataset_view(self, email, view_id):
        collection = self.db.views
        doc = collection.find_one(dict(_id=ObjectId(view_id)))
        self.get_dataset(email, doc['dataset_id'])
        return format_doc(doc)

    def upsert_dataset_view(self, email, dataset_id, view):
        if not self.capabilities()[SERVER_CAPABILITY_LINKS]:
            return
        self.get_dataset(email, dataset_id)
        collection = self.db.views
        view['last_updated'] = datetime.datetime.utcnow()

        if email is not None:
            view['email'] = email
        if dataset_id is not None:
            view['dataset_id'] = dataset_id
        if view.get('id') is None:
            return dict(id=str(collection.insert_one(view).inserted_id), last_updated=view['last_updated'])
        else:
            view_id = view.pop('id')
            collection.update_one(dict(_id=ObjectId(view_id)), {'$set': view})
            return dict(id=view_id, last_updated=view['last_updated'])

    def is_importer(self, email):
        # TODO check if user can modify dataset
        user = self.db.users.find_one(dict(email=email))
        if 'importer' not in user:
            raise False
        return user['importer']

    def delete_dataset(self, email, dataset_id):
        if not self.capabilities()[SERVER_CAPABILITY_DELETE_DATASET]:
            return

        self.get_dataset(email, dataset_id, True)
        self.db.datasets.delete_one(dict(_id=ObjectId(dataset_id)))
        self.db.filters.delete_many(dict(dataset_id=dataset_id))
        self.db.views.delete_many(dict(dataset_id=dataset_id))
        self.db.categories.delete_many(dict(dataset_id=dataset_id))
        self.db.feature_sets.delete_many(dict(dataset_id=dataset_id))

        for doc in self.db.jobs.find(dict(dataset_id=dataset_id), dict(result=1)):
            result = doc.get('result')
            self.delete_job_result(result)
        self.db.jobs.delete_many(dict(dataset_id=dataset_id))

    def upsert_dataset(self, email, readers, dataset):

        if dataset.get('id') is None and not self.capabilities()[SERVER_CAPABILITY_ADD_DATASET]:
            return
        if dataset.get('id') is not None and not self.capabilities()[SERVER_CAPABILITY_EDIT_DATASET]:
            return
        collection = self.db.datasets
        if readers is not None:
            readers = set(readers)
            if email in readers:
                readers.remove(email)
            readers.add(email)
            dataset['readers'] = list(readers)
        dataset['last_updated'] = datetime.datetime.utcnow()
        if dataset.get('id') is None:  # new dataset
            if email != '':
                user = self.db.users.find_one(dict(email=email))
                if 'importer' not in user or not user['importer']:
                    raise InvalidUsage('Not authorized', 403)
            dataset['owners'] = [email]
            if 'readers' not in dataset:
                dataset['readers'] = [email]
            return str(collection.insert_one(dataset).inserted_id)
        else:  # update
            self.get_dataset(email, dataset['id'], True)
            dataset_id = dataset.pop('id')
            collection.update_one(dict(_id=ObjectId(dataset_id)), {'$set': dataset})
            return dataset_id

    def get_feature_sets(self, email, dataset_id):
        self.get_dataset(email, dataset_id)
        collection = self.db.feature_sets
        results = []
        for doc in collection.find(dict(dataset_id=dataset_id)):
            results.append(format_doc(doc))
        return results

    def delete_feature_set(self, email, dataset_id, set_id):
        if not self.capabilities()[SERVER_CAPABILITY_FEATURE_SETS]:
            return
        collection = self.db.feature_sets
        doc = collection.find_one(dict(_id=ObjectId(set_id)))
        self.get_dataset(email, doc['dataset_id'])
        collection.delete_one(dict(_id=ObjectId(set_id)))

    def upsert_feature_set(self, email, dataset_id, set_id, category, name, features):
        if not self.capabilities()[SERVER_CAPABILITY_FEATURE_SETS]:
            return
        self.get_dataset(email, dataset_id)
        collection = self.db.feature_sets
        entity_update = {}
        entity_update['last_updated'] = datetime.datetime.utcnow()
        if name is not None:
            entity_update['name'] = name
        if features is not None:
            entity_update['features'] = features
        if email is not None:
            entity_update['email'] = email
        if dataset_id is not None:
            entity_update['dataset_id'] = dataset_id
        if category is not None:
            entity_update['category'] = category
        if set_id is None:
            return str(collection.insert_one(entity_update).inserted_id)
        else:
            collection.update_one(dict(_id=ObjectId(set_id)), {'$set': entity_update})
            return set_id

    def get_gridfs(self):
        import gridfs
        if self.fs is None:
            self.fs = gridfs.GridFS(self.db)
        return self.fs

    def create_job(self, email, dataset_id, job_name, job_type, params):
        if not self.capabilities()[SERVER_CAPABILITY_JOBS]:
            return
        self.get_dataset(email, dataset_id)
        collection = self.db.jobs
        job_id = str(collection.insert_one(
            dict(dataset_id=dataset_id, status='pending', name=job_name, email=email, type=job_type, params=params,
                 submitted=datetime.datetime.utcnow())).inserted_id)
        return job_id

    def get_job(self, email, job_id, return_type):
        collection = self.db.jobs
        doc = collection.find_one(dict(_id=ObjectId(job_id)))
        if doc is None:
            return None
        self.get_dataset(email, doc['dataset_id'])
        if return_type == 'result':
            result = doc.get('result')
            if isinstance(result, dict) and result.get('url') is not None:
                return result
            else:
                fs = self.get_gridfs()
                return fs.get(ObjectId(doc['result'])).read()
        elif return_type == 'status':
            return dict(status=doc['status'])
        elif return_type == 'params':
            return dict(params=doc.get('params'))

    def get_jobs(self, email, dataset_id):
        self.get_dataset(email, dataset_id)
        collection = self.db.jobs
        results = []
        for doc in collection.find(dict(dataset_id=dataset_id),
                                   dict(name=1, status=1, email=1, type=1, submitted=1, readonly=1)):
            results.append(
                dict(id=str(doc['_id']), name=doc['name'], status=doc['status'], type=doc['type'], email=doc['email'],
                     submitted=doc.get('submitted'), readonly=doc.get('readonly', False)))
        return results

    def annotate_job(self, email, job_id, annotations):
        collection = self.db.jobs
        doc = collection.find_one(dict(_id=ObjectId(job_id)))
        self.get_dataset(email, doc['dataset_id'])
        collection.update_one(dict(_id=ObjectId(job_id)),
                              {'$set': dict(annotations=annotations, last_updated=datetime.datetime.utcnow())})

    def update_job(self, email, job_id, status, result):
        if not self.capabilities()[SERVER_CAPABILITY_JOBS]:
            return
        collection = self.db.jobs
        doc = collection.find_one(dict(_id=ObjectId(job_id)))
        self.get_dataset(email, doc['dataset_id'])
        if doc.get('readonly', False):
            raise InvalidUsage('Not authorized', 403)
        if result is not None:
            if os.environ.get(CIRRO_JOB_RESULTS) is not None:  # save to directory
                result = save_job_result_to_file(result, job_id)
            else:
                result = ujson.dumps(result, double_precision=2, orient='values')
                result = str(self.get_gridfs().put(result, encoding='ascii'))

        collection.update_one(dict(_id=ObjectId(job_id)), {'$set': dict(status=status, result=result)})

    def delete_job_result(self, result):
        if result is not None:
            if isinstance(result, dict) and result.get('url') is not None:
                fs = get_fs(result['url'])
                if fs.exists(result['url']):
                    fs.rm(result['url'], recursive=True)
            else:
                self.get_gridfs().delete(ObjectId(result))

    def delete_job(self, email, job_id):
        if not self.capabilities()[SERVER_CAPABILITY_JOBS]:
            return
        collection = self.db.jobs
        doc = collection.find_one(dict(_id=ObjectId(job_id)), dict(email=1))
        if doc['email'] == email and not doc.get('readonly', False):
            collection.delete_one(dict(_id=ObjectId(job_id)))
            result = doc.get('result')
            self.delete_job_result(result)
        else:
            raise InvalidUsage('Not authorized', 403)



import anndata
import numpy as np
import pandas as pd

DATA_TYPE_MODULE = 'module'
DATA_TYPE_UNS_KEY = 'data_type'
ADATA_MODULE_UNS_KEY = 'anndata_module'


def get_base(adata):
    base = None
    if 'log1p' in adata.uns and adata.uns['log1p']['base'] is not None:
        base = adata.uns['log1p'][base]
    return base


def adata_to_df(adata):
    df = pd.DataFrame(adata.X, index=adata.obs.index, columns=adata.var.index)
    for key in adata.layers.keys():
        df2 = pd.DataFrame(adata.layers[key], index=adata.obs.index.astype(str) + '-{}'.format(key),
                           columns=adata.var.index)
        df = pd.concat((df, df2), axis=0)

    df = df.T.join(adata.var)
    df.index.name = 'id'
    return df.reset_index()


def get_scanpy_marker_keys(dataset):
    marker_keys = []
    try:
        dataset.uns  # dataset can be AnnData or zarr group
    except AttributeError:
        return marker_keys
    from collections.abc import Mapping
    for key in dataset.uns.keys():
        rank_genes_groups = dataset.uns[key]
        if isinstance(rank_genes_groups, Mapping) and 'names' in rank_genes_groups and (
                'pvals' in rank_genes_groups or 'pvals_adj' in rank_genes_groups or 'scores' in rank_genes_groups) and len(
            rank_genes_groups['names'][0]) > 0 and not isinstance(rank_genes_groups['names'][0][0], bytes):
            marker_keys.append(key)
    return marker_keys


def get_pegasus_marker_keys(dataset):
    marker_keys = []
    try:
        dataset.varm  # dataset can be AnnData or zarr group
    except AttributeError:
        return marker_keys
    for key in dataset.varm.keys():
        d = dataset.varm[key]
        if isinstance(d, np.recarray):
            try:
                ''.join(d.dtype.names).index('log2Mean')
                marker_keys.append(key)
            except ValueError:
                pass
    return marker_keys


def obs_stats(adata, columns):
    df = adata.obs[columns]
    # variables on columns, stats on rows, transpose so that stats are on columns
    return df.agg(['min', 'max', 'sum', 'mean']).T


def X_stats(adata):
    X = adata.X
    return pd.DataFrame(
        data={'min': X.min(axis=0).toarray().flatten(),
              'max': X.max(axis=0).toarray().flatten(),
              'sum': X.sum(axis=0).flatten(),
              'numExpressed': X.getnnz(axis=0),
              'mean': X.mean(axis=0)}, index=adata.var.index)


def dataset_schema(dataset, n_features=10):
    """ Gets dataset schema.

    Returns
        schema dict. Example:
        {"version":"1.0.0",
        "categoryOrder":{
            "louvain":["0","1","2","3","4","5","6","7"],
            "leiden":["0","1","2","3","4","5","6","7"]},
        "var":["TNFRSF4","CPSF3L","ATAD3C"],
        "obs":["percent_mito","n_counts"],
        "obsCat":["louvain","leiden"],
        "shape":[2638,1838],
        "embeddings":[{"name":"X_pca","dimensions":3},{"name":"X_pca","dimensions":2},{"name":"X_umap","dimensions":2}]
    }
    """
    obs_cat = []
    obs = []
    marker_results = []
    de_results_format = 'records'

    prior_marker_results = dataset.uns.get('markers', [])
    if isinstance(prior_marker_results, str):
        import json
        prior_marker_results = json.loads(prior_marker_results)
    marker_results += prior_marker_results

    de_results = []  # array of dicts containing params logfoldchanges, pvals_adj, scores, names
    category_to_order = {}
    embeddings = []
    field_to_value_to_color = dict()  # field -> value -> color

    for scanpy_marker_key in get_scanpy_marker_keys(dataset):
        rank_genes_groups = dataset.uns[scanpy_marker_key]
        has_fc = 'logfoldchanges' in rank_genes_groups
        min_fold_change = 1
        params = rank_genes_groups['params']
        if not isinstance(params, dict):
            from anndata._io.zarr import read_attribute
            params = {k: read_attribute(params[k]) for k in params.keys()}

        # pts and pts_rest in later scanpy versions
        rank_genes_groups_keys = list(rank_genes_groups.keys())
        for k in ['params', 'names']:
            if k in rank_genes_groups_keys:
                rank_genes_groups_keys.remove(k)
        if 'pvals' in rank_genes_groups_keys and 'pvals_adj' in rank_genes_groups_keys:
            rank_genes_groups_keys.remove('pvals')
        category = '{} ({})'.format(params['groupby'], scanpy_marker_key)
        de_result_name = category
        de_result_df = None
        group_names = rank_genes_groups['names'].dtype.names
        de_result = dict(id='cirro-{}'.format(scanpy_marker_key),
                         type='de',
                         readonly=True,
                         groups=group_names,
                         fields=rank_genes_groups_keys,
                         name=de_result_name)
        for group_name in group_names:
            group_df = pd.DataFrame(index=rank_genes_groups['names'][group_name][...])
            group_df = group_df[group_df.index != 'nan']
            for rank_genes_groups_key in rank_genes_groups_keys:
                values = rank_genes_groups[rank_genes_groups_key][group_name][...]
                column_name = '{}:{}'.format(group_name, rank_genes_groups_key)
                group_df[column_name] = values

            if de_result_df is None:
                de_result_df = group_df
            else:
                de_result_df = de_result_df.join(group_df, how='outer')
            if n_features > 0:
                markers_df = group_df
                if has_fc:
                    markers_df = group_df[group_df['{}:logfoldchanges'.format(group_name)] > min_fold_change]
                if len(markers_df) > 0:
                    marker_results.append(
                        dict(category=de_result_name, name=str(group_name),
                             features=markers_df.index[:n_features]))

        if de_results_format == 'records':
            de_result_data = de_result_df.reset_index().to_dict(orient='records')
        else:
            de_result_data = dict(index=de_result_df.index[...])
            for c in de_result_df:
                de_result_data[c] = de_result_df[c]

        de_result['params'] = params
        de_result['data'] = de_result_data
        de_results.append(de_result)
    for pg_marker_key in get_pegasus_marker_keys(dataset):
        de_res = dataset.varm[pg_marker_key]
        key_name = pg_marker_key
        if pg_marker_key.startswith('rank_genes_'):
            key_name = pg_marker_key[len('rank_genes_'):]
        names = de_res.dtype.names
        field_names = set()  # e.g. 1:auroc
        group_names = set()
        for name in names:
            index = name.rindex(':')
            field_name = name[index + 1:]
            group_name = name[:index]
            field_names.add(field_name)
            group_names.add(group_name)
        group_names = list(group_names)
        field_names = list(field_names)

        de_result_df = pd.DataFrame(data=de_res, index=dataset.var.index)
        de_result_df.index.name = 'index'
        if de_results_format == 'records':
            de_result_data = de_result_df.reset_index().to_dict(orient='records')
        else:
            de_result_data = dict(index=de_result_df.index)
            for c in de_res:
                de_result_data[c] = de_result_df[c]
        de_result = dict(id='cirro-{}'.format(pg_marker_key), type='de', name=key_name,
                         color='log2FC' if 'log2FC' in field_names else field_names[0],
                         size='mwu_qval' if 'mwu_qval' in field_names else field_names[0],
                         groups=group_names,
                         fields=field_names)
        de_result['data'] = de_result_data
        de_results.append(de_result)

        if n_features > 0:
            field_use = None
            for field in ['mwu_qval', 'auroc', 't_qval']:
                if field in field_names:
                    field_use = field
                    break
            if field_use is not None:
                field_ascending = field_use != 'auroc'
                for group_name in group_names:
                    fc_column = '{}:log2FC'.format(group_name)
                    name = '{}:{}'.format(group_name, field_name)

                    idx_up = de_result_df[fc_column] > 0
                    df_up = de_result_df.loc[idx_up].sort_values(by=[name, fc_column],
                                                                 ascending=[field_ascending, False])
                    features = df_up[:n_features].index.values

                    marker_results.append(dict(category=key_name, name=str(group_name), features=features))

    categories_node = dataset.obs['__categories'] if '__categories' in dataset.obs else None
    for key in dataset.obs.keys():
        if categories_node is not None and (key == '__categories' or key == 'index'):
            continue
        val = dataset.obs[key]
        if categories_node is not None and key in categories_node:
            categories = categories_node[key][...]
            ordered = categories_node[key].attrs.get('ordered', False)
            val = pd.Categorical.from_codes(val[...], categories, ordered=ordered)

        if pd.api.types.is_categorical_dtype(val) or pd.api.types.is_bool_dtype(
                val) or pd.api.types.is_object_dtype(val):
            obs_cat.append(key)
        else:
            obs.append(key)
        if pd.api.types.is_categorical_dtype(val):
            categories = val.cat.categories
            if len(categories) < 100:  # preserve order
                category_to_order[key] = dataset.obs[key].cat.categories

            color_field = key + '_colors'
            if color_field in dataset.uns:
                colors = dataset.uns[color_field][...]
                if len(categories) == len(colors):
                    color_map = dict()
                    for j in range(len(categories)):
                        color_map[str(categories[j])] = colors[j]
                    field_to_value_to_color[key] = color_map

    # spatial_node = adata.uns['spatial'] if 'spatial' in adata.uns else None
    #
    # if spatial_node is not None:
    #     spatial_node_keys = list(spatial_node.keys())  # list of datasets
    #     if len(spatial_node_keys) == 1:
    #         spatial_node = spatial_node[spatial_node_keys[0]]

    images_node = dataset.uns.get('images',
                                  [])  # list of {type:image or meta_image, name:image name, image:path to image, spot_diameter:Number}
    image_names = list(map(lambda x: x['name'], images_node))
    layers = []
    try:
        dataset.layers  # dataset can be AnnData or zarr group
        layers = list(dataset.layers.keys())
    except AttributeError:
        pass

    for key in dataset.obsm.keys():
        dim = dataset.obsm[key].shape[1]
        if 1 < dim <= 3:
            embedding = dict(name=key, dimensions=dim)
            if dim == 2:
                try:
                    image_index = image_names.index(key)
                    embedding['spatial'] = images_node[image_index]
                except ValueError:
                    pass

            embeddings.append(embedding)
    meta_images = dataset.uns.get('meta_images', [])
    for meta_image in meta_images:
        embeddings.append(meta_image)
    schema_dict = {'version': '1.0.0'}
    schema_dict['results'] = de_results
    schema_dict['colors'] = field_to_value_to_color
    schema_dict['markers'] = marker_results
    schema_dict['embeddings'] = embeddings
    schema_dict['categoryOrder'] = category_to_order
    schema_dict['layers'] = layers
    var_df = dataset.var
    if not isinstance(var_df, pd.DataFrame):
        from anndata._io.zarr import read_attribute
        var_df = read_attribute(dataset.var)
    var_df.index.name = 'id'
    schema_dict['var'] = var_df.reset_index().to_dict(orient='records')
    modules_df = None
    if ADATA_MODULE_UNS_KEY in dataset.uns and isinstance(dataset.uns[ADATA_MODULE_UNS_KEY], anndata.AnnData):
        modules_df = dataset.uns[ADATA_MODULE_UNS_KEY].var
        #  if not isinstance(module_var, pd.DataFrame):
        #             from anndata._io.zarr import read_attribute
        #             module_var = read_attribute(module_var)
    if modules_df is not None:
        modules_df.index.name = 'id'
        schema_dict['modules'] = modules_df.reset_index().to_dict(orient='records')

    schema_dict['obs'] = obs
    schema_dict['obsCat'] = obs_cat
    shape = dataset.shape if isinstance(dataset, anndata.AnnData) else dataset.X.attrs.shape
    schema_dict['shape'] = shape
    return schema_dict



import argparse
import logging
import os

import anndata
import numpy as np
import pandas as pd
import scipy.sparse
from cirrocumulus.anndata_util import get_scanpy_marker_keys, dataset_schema, ADATA_MODULE_UNS_KEY
from cirrocumulus.io_util import get_markers, filter_markers, add_spatial, SPATIAL_HELP, unique_id
from cirrocumulus.util import to_json, get_fs, open_file

logger = logging.getLogger("cirro")

cluster_fields = ['anno', 'cell_type', 'celltype', 'leiden', 'louvain', 'seurat_cluster', 'cluster']
categorical_fields_convert = ['seurat_clusters']


def read_adata(path, spatial_directory=None, use_raw=False):
    if path.lower().endswith('.loom'):
        adata = anndata.read_loom(path)
    elif path.lower().endswith('.zarr'):
        adata = anndata.read_zarr(path)
    else:
        adata = anndata.read(path)
    if 'module' in adata.uns:
        adata.uns[ADATA_MODULE_UNS_KEY] = anndata.AnnData(X=adata.uns['module']['X'],
                                                          var=adata.uns['module']['var'])
    if use_raw and adata.raw is not None and adata.shape[0] == adata.raw.shape[0]:
        logger.info('Using adata.raw')
        adata = anndata.AnnData(X=adata.raw.X, var=adata.raw.var, obs=adata.obs, obsm=adata.obsm, uns=adata.uns)

    if spatial_directory is not None:
        if not add_spatial(adata, spatial_directory):
            logger.info('No spatial data found in {}'.format(spatial_directory))

    for field in categorical_fields_convert:
        if field in adata.obs and not pd.api.types.is_categorical_dtype(adata.obs[field]):
            logger.info('Converting {} to categorical'.format(field))
            adata.obs[field] = adata.obs[field].astype(str).astype('category')
    return adata


class PrepareData:

    def __init__(self, datasets, output, dimensions=None, groups=[], group_nfeatures=10, markers=[],
                 output_format='parquet', no_auto_groups=False, save_whitelist=None):
        self.groups = groups
        self.group_nfeatures = group_nfeatures
        self.markers = markers
        self.output_format = output_format
        self.no_auto_groups = no_auto_groups
        self.save_whitelist = save_whitelist

        for dataset in datasets:
            for key in list(dataset.obsm.keys()):
                m = dataset.obsm[key]
                dim = m.shape[1]
                if not (1 < dim <= 3):
                    del dataset.obsm[key]
        if len(datasets) > 1:
            for i in range(len(datasets)):
                dataset = datasets[i]
                if 'group' not in dataset.var:
                    dataset.var['group'] = dataset.uns.get('name', 'dataset {}'.format(i + 1))
                if i > 0 and not np.array_equal(datasets[0].obs.index, dataset.obs.index):
                    raise ValueError('obs ids are not equal')
            dataset = anndata.concat(datasets, axis=1, label='group', merge='unique')
        else:
            dataset = datasets[0]
        dataset.var.index = dataset.var.index.str.replace('/', '_')
        dataset.var_names_make_unique()
        dataset.obs.index.name = 'id'
        dataset.var.index.name = 'id'
        self.base_output = output
        dimensions_supplied = dimensions is not None and len(dimensions) > 0
        self.dimensions = [] if not dimensions_supplied else dimensions
        self.measures = []
        self.others = []
        self.dataset = dataset
        if scipy.sparse.issparse(dataset.X) and not scipy.sparse.isspmatrix_csc(dataset.X):
            dataset.X = dataset.X.tocsc()
        for i in range(len(dataset.obs.columns)):
            name = dataset.obs.columns[i]
            c = dataset.obs[name]
            if pd.api.types.is_object_dtype(c):
                dataset.obs[name] = dataset.obs[name].astype('category')
                c = dataset.obs[name]
            if not dimensions_supplied and pd.api.types.is_categorical_dtype(c):
                if 1 < len(c.cat.categories) < 2000:
                    self.dimensions.append(name)
                    if c.isna().sum() > 0:
                        logger.info('Replacing nans in {}'.format(name))
                        dataset.obs[name] = dataset.obs[name].astype(str)
                        dataset.obs.loc[dataset.obs[name].isna(), name] = ''
                        dataset.obs[name] = dataset.obs[name].astype('category')
                else:
                    self.others.append(name)
            elif not pd.api.types.is_string_dtype(c) and not pd.api.types.is_object_dtype(c):
                self.measures.append('obs/' + name)
            else:
                self.others.append(name)

    def execute(self):
        output_format = self.output_format
        dataset = self.dataset
        if self.groups is None and not self.no_auto_groups:
            groups = []
            existing_fields = set()
            scanpy_marker_keys = get_scanpy_marker_keys(dataset)
            for key in scanpy_marker_keys:
                group_by = dataset.uns[key]['params']['groupby']
                if isinstance(group_by, np.ndarray):
                    group_by = ','.join(group_by)
                existing_fields.add(group_by)
            for field in dataset.obs.columns:
                field_lc = field.lower()
                for cluster_field in cluster_fields:
                    if field_lc.find(cluster_field) != -1 and cluster_field not in existing_fields:
                        groups.append(field)
                        break
            self.groups = groups
        if self.groups is not None and len(self.groups) > 0:
            use_pegasus = False
            use_scanpy = False
            try:
                import pegasus as pg
                use_pegasus = True
            except ModuleNotFoundError:
                pass
            if not use_pegasus:
                try:
                    import scanpy as sc
                    use_scanpy = True
                except ModuleNotFoundError:
                    pass
            if not use_pegasus and not use_scanpy:
                raise ValueError('Please install pegasuspy or scanpy to compute markers')

            for group in self.groups:
                field = group
                if group not in dataset.obs:  # test if multiple comma separated fields
                    split_groups = group.split(',')
                    if len(split_groups) > 1:
                        use_split_groups = True
                        for split_group in split_groups:
                            if split_group not in dataset.obs:
                                use_split_groups = False
                                break
                        if use_split_groups:
                            dataset.obs[field] = dataset.obs[split_groups[0]].str.cat(dataset.obs[split_groups[1:]],
                                                                                      sep=',')

                if field in dataset.obs:
                    if not pd.api.types.is_categorical_dtype(dataset.obs[field]):
                        dataset.obs[field] = dataset.obs[field].astype(str).astype('category')
                    if len(dataset.obs[field].cat.categories) > 1:
                        logger.info('Computing markers for {}'.format(field))
                        key_added = 'rank_genes_' + str(field)
                        if use_pegasus:
                            pg.de_analysis(dataset, cluster=field, de_key=key_added)
                        else:
                            sc.tl.rank_genes_groups(dataset, field, key_added=key_added, method='t-test')
                else:
                    logger.info(group + ' not found in ' + ', '.join(dataset.obs.columns))
        schema = self.get_schema()
        schema['format'] = output_format
        if output_format in ['parquet', 'zarr']:
            output_dir = self.base_output
        else:
            output_dir = os.path.splitext(self.base_output)[0]
        filesystem = get_fs(output_dir)
        filesystem.makedirs(output_dir, exist_ok=True)
        results = schema.get('results', [])

        if len(results) > 0:
            uns_dir = os.path.join(output_dir, 'uns')
            is_gzip = output_format != 'jsonl'
            filesystem.makedirs(uns_dir, exist_ok=True)

            for i in range(len(results)):
                full_result = results[i]
                result_id = full_result.pop('id')
                # keep id, name, type in schema, store rest externally
                results[i] = dict(id=result_id, name=full_result.pop('name'), type=full_result.pop('type'),
                                  content_type='application/json', content_encoding='gzip' if is_gzip else None)
                json_result = to_json(full_result)

                result_path = os.path.join(uns_dir, result_id + '.json.gz') if is_gzip else os.path.join(uns_dir,
                                                                                                             result_id + '.json')
                with open_file(result_path, 'wt', compression='gzip' if is_gzip else None) as out:
                    out.write(json_result)
        images = dataset.uns.get('images')
        if images is not None:
            image_dir = os.path.join(output_dir, 'images')
            filesystem.makedirs(image_dir, exist_ok=True)
            for image in images:
                src = image['image']
                dest = os.path.join(image_dir, os.path.basename(src))
                filesystem.copy(src, dest)
                image['image'] = 'images/' + os.path.basename(src)

        if output_format == 'parquet':
            from cirrocumulus.parquet_output import save_dataset_pq
            save_dataset_pq(dataset, schema, self.base_output, filesystem, self.save_whitelist)
        elif output_format == 'jsonl':
            from cirrocumulus.jsonl_io import save_dataset_jsonl
            save_dataset_jsonl(dataset, schema, output_dir, self.base_output, filesystem)
        elif output_format == 'zarr':
            from cirrocumulus.zarr_output import save_dataset_zarr
            save_dataset_zarr(dataset, schema, self.base_output, filesystem, self.save_whitelist)
        else:
            raise ValueError("Unknown format")

    def get_schema(self):
        result = dataset_schema(self.dataset, n_features=self.group_nfeatures)
        markers = result.get('markers', [])

        if self.markers is not None:  # add results specified from file
            markers += get_markers(self.markers)
            markers = filter_markers(self.dataset, markers)

        for marker in markers:
            if marker.get('id') is None:
                marker['id'] = unique_id()
            marker['readonly'] = True
        result['markers'] = markers
        result['format'] = self.output_format
        return result


def main(argsv):
    parser = argparse.ArgumentParser(
        description='Prepare a dataset for cirrocumulus server.')
    parser.add_argument('dataset', help='Path to a h5ad, loom, or Seurat file', nargs='+')
    parser.add_argument('--out', help='Path to output directory')
    parser.add_argument('--format', help='Output format', choices=['parquet', 'jsonl', 'zarr'],
                        default='zarr')
    parser.add_argument('--whitelist',
                        help='Optional whitelist of fields to save. Only applies when output format is parquet',
                        choices=['obs', 'obsm', 'X'],
                        action='append')
    parser.add_argument('--markers',
                        help='Path to JSON file of precomputed markers that maps name to features. For example {"a":["gene1", "gene2"], "b":["gene3"]',
                        action='append')
    parser.add_argument('--no-auto-groups', dest='no_auto_groups',
                        help='Disable automatic cluster field detection to compute differential expression results for',
                        action='store_true')
    parser.add_argument('--groups',
                        help='List of groups to compute markers for (e.g. louvain). Markers created with cumulus/scanpy are automatically included. Separate multiple groups with a comma to combine groups using "AND" logic (e.g. louvain,day)',
                        action='append')
    parser.add_argument('--group_nfeatures', help='Number of marker genes/features to include', type=int, default=10)
    parser.add_argument('--spatial', help=SPATIAL_HELP)
    args = parser.parse_args(argsv)
    logger.setLevel(logging.INFO)
    logger.addHandler(logging.StreamHandler())
    out = args.out
    no_auto_groups = args.no_auto_groups
    save_whitelist = args.whitelist
    input_datasets = args.dataset  # multimodal
    output_format = args.format
    if out is None:
        out = os.path.splitext(os.path.basename(input_datasets[0]))[0]
    if out.endswith('/'):
        out = out[:len(out) - 1]
    output_format2extension = dict(parquet='.cpq', jsonl='.jsonl', zarr='.zarr', h5ad='.h5ad')
    if not out.lower().endswith(output_format2extension[output_format]):
        out += output_format2extension[output_format]

    datasets = []
    tmp_files = []
    for input_dataset in input_datasets:
        use_raw = False
        if input_dataset.lower().endswith('.rds'):
            import subprocess
            import tempfile
            import pkg_resources

            _, h5_file = tempfile.mkstemp(suffix='.h5ad')
            os.remove(h5_file)
            subprocess.check_call(
                ['Rscript', pkg_resources.resource_filename("cirrocumulus", 'seurat2h5ad.R'), input_dataset, h5_file])
            input_dataset = h5_file
            tmp_file = h5_file
            use_raw = True
            tmp_files.append(tmp_file)
        adata = read_adata(input_dataset, spatial_directory=args.spatial, use_raw=use_raw)
        datasets.append(adata)
        adata.uns['name'] = os.path.splitext(os.path.basename(input_dataset.rstrip('/')))[0]

    prepare_data = PrepareData(datasets=datasets, output=out, dimensions=args.groups, groups=args.groups,
                               group_nfeatures=args.group_nfeatures,
                               markers=args.markers, output_format=output_format, no_auto_groups=no_auto_groups,
                               save_whitelist=save_whitelist)
    prepare_data.execute()
    for tmp_file in tmp_files:
        os.remove(tmp_file)


if __name__ == '__main__':
    import sys

    main(sys.argv)



import os

import anndata

from cirrocumulus.anndata_dataset import AnndataDataset
from cirrocumulus.envir import CIRRO_DATABASE, CIRRO_AUTH, CIRRO_JOB_TYPE, \
    CIRRO_JOB_RESULTS, CIRRO_CELL_ONTOLOGY
from cirrocumulus.io_util import get_markers, filter_markers, add_spatial, \
    SPATIAL_HELP
from cirrocumulus.local_db_api import LocalDbAPI
from cirrocumulus.util import get_fs


def configure_app(app, list_of_dataset_paths, spatial_directories,
                  marker_paths):
    from cirrocumulus.api import dataset_api
    from cirrocumulus.no_auth import NoAuth
    try:
        from cirrocumulus.parquet_dataset import ParquetDataset
        dataset_api.add(ParquetDataset())
    except ModuleNotFoundError:
        pass
    try:
        from cirrocumulus.zarr_dataset import ZarrDataset
        dataset_api.add(ZarrDataset())
    except ModuleNotFoundError:
        pass
    app.config[CIRRO_AUTH] = NoAuth()
    os.environ[CIRRO_JOB_TYPE + 'de'] = 'cirrocumulus.job_api.run_de'
    anndata_dataset = AnndataDataset()
    dataset_ids = []
    for dataset_paths in list_of_dataset_paths:
        dataset_paths = dataset_paths.split(',')
        dataset_id = dataset_paths[0]
        dataset_ids.append(dataset_id)
        if len(dataset_paths) > 1:
            datasets = []
            for i in range(len(dataset_paths)):
                dataset = anndata_dataset.get_data(get_fs(dataset_paths[i]),
                                                   dataset_paths[i])
                if 'group' not in dataset.var:
                    dataset.var['group'] = dataset.uns.get('name',
                                                           'dataset {}'.format(
                                                               i + 1))
                datasets.append(dataset)
            adata = anndata.concat(datasets, axis=1, label='group',
                                   merge='unique')
            dataset.obsm = datasets[0].obsm
            adata.var.index = adata.var.index.str.replace('/', '_')
            adata.var_names_make_unique()
            anndata_dataset.add_data(dataset_id, adata)
        dataset_api.add(anndata_dataset)

    app.config[CIRRO_DATABASE] = LocalDbAPI(dataset_ids)

    if spatial_directories is not None and len(spatial_directories) > 0:
        for i in range(len(spatial_directories)):
            spatial_directory = spatial_directories[i]
            if spatial_directory != '':
                adata = anndata_dataset.get_data(get_fs(dataset_ids[i]),
                                                 dataset_ids[i])
                if not add_spatial(adata, spatial_directory):
                    print(
                        'No spatial data found in {}'.format(spatial_directory))

    if marker_paths is not None and len(marker_paths) > 0:
        markers = get_markers(marker_paths)
        for dataset_id in dataset_ids:
            d = anndata_dataset.get_data(get_fs(dataset_id), dataset_id)
            existing_markers = d.uns.get('markers', [])
            markers += existing_markers
            # remove genes in dict that are not in dataset
            d.uns['markers'] = filter_markers(d, markers)


def create_app():
    from cirrocumulus.api import cirro_blueprint
    from flask_compress import Compress
    from flask import Flask, send_from_directory
    app = Flask(__name__, static_folder='client', static_url_path='')
    app.config['SEND_FILE_MAX_AGE_DEFAULT'] = 0
    app.register_blueprint(cirro_blueprint, url_prefix='/api')

    @app.route('/')
    def root():
        return send_from_directory(
            os.path.abspath(os.path.join(app.root_path, "client")),
            "index.html")

    Compress(app)
    return app


def main(argsv):
    import argparse

    parser = argparse.ArgumentParser(description='Run cirrocumulus')
    parser.add_argument('dataset',
                        help='Path to dataset in h5ad, loom, Seurat, TileDB, zarr, or STAR-Fusion format. Separate multiple datasets with '
                             'a comma instead of a space in order to join datasets by cell id',
                        nargs='+')
    parser.add_argument('--spatial', help=SPATIAL_HELP, nargs='*')
    parser.add_argument('--markers',
                        help='Path to JSON file that maps name to features. For example {"a":["gene1", "gene2"], "b":["gene3"]}',
                        nargs='*')
    parser.add_argument('--host',
                        help='Host IP address')  # set to 0.0.0.0 to make it accessible from other computers WITHOUT SECURITY.

    parser.add_argument('--port', help='Server port', default=5000, type=int)
    parser.add_argument('--no-open', dest='no_open',
                        help='Do not open your web browser',
                        action='store_true')
    parser.add_argument('--results',
                        help='URL to save user computed results (e.g. differential expression)')
    parser.add_argument('--ontology',
                        help='Path to ontology in OBO format for annotation')
    args = parser.parse_args(argsv)
    if args.results is not None:
        os.environ[CIRRO_JOB_RESULTS] = args.results
    else:
        os.environ[CIRRO_JOB_RESULTS] = os.path.join(
            os.path.dirname(args.dataset[0].rstrip('/')), 'results')
    get_fs(os.environ[CIRRO_JOB_RESULTS]).makedirs(
        os.environ[CIRRO_JOB_RESULTS], exist_ok=True)
    if args.ontology is not None:
        os.environ[CIRRO_CELL_ONTOLOGY] = args.ontology
    app = create_app()
    configure_app(app, args.dataset, args.spatial, args.markers)
    if not args.no_open:
        import webbrowser, requests
        host = args.host if args.host is not None else 'http://127.0.0.1'
        url = host + ':' + str(args.port)
        try:
            if requests.get(url).ok:
                import sys
                sys.exit('Address already in use')
        except:
            pass
        webbrowser.open(url)
    app.run(host=args.host, port=args.port, debug=False)


if __name__ == "__main__":
    import sys

    main(sys.argv[1:])



import json
import os
from abc import abstractmethod, ABC

import pandas as pd
from cirrocumulus.api import get_file_path


class AbstractDataset(ABC):

    def __init__(self):
        super().__init__()

    @abstractmethod
    def read_dataset(self, filesystem, path, keys, dataset):
        pass

    @abstractmethod
    def get_suffixes(self):
        pass

    def get_result(self, filesystem, path, dataset, result_id):
        return get_file_path(os.path.join('uns', result_id + '.json.gz'), path)

    def get_dataset_info(self, filesystem, path):
        """Returns a dict with shape, var, modules
       """
        s = self.get_schema(filesystem, path)
        d = dict()
        var = s['var']
        d['var'] = pd.Index(map(lambda x: x['id'], var)) if isinstance(var[0], dict) else pd.Index(var)
        d['shape'] = s['shape']
        return d

    def get_schema(self, filesystem, path):
        if path.endswith('.gz'):
            import gzip
            with gzip.open(filesystem.open(path)) as s:
                return json.load(s)
        else:
            with filesystem.open(path) as s:
                return json.load(s)



import os

from cirrocumulus.envir import CIRRO_AUTH_CLIENT_ID, CIRRO_DB_URI, CIRRO_SERVE, \
    CIRRO_FOOTER, CIRRO_UPLOAD, CIRRO_BRAND, CIRRO_DATABASE_CLASS, CIRRO_JOB_RESULTS, CIRRO_AUTH, CIRRO_DATABASE, \
    CIRRO_DATASET_PROVIDERS, CIRRO_JOB_TYPE, CIRRO_CELL_ONTOLOGY, CIRRO_AUTH_BASE_URL
from cirrocumulus.launch import create_app
from cirrocumulus.util import create_instance, add_dataset_providers, get_fs

app = None

DEFAULT_DB_URI = 'mongodb://localhost:27017/cirrocumulus'


def cached_app():
    global app
    if app is None:
        app = create_app()
        configure_app(app)
    return app


def configure_app(app):
    from cirrocumulus.no_auth import NoAuth
    auth_client_id = os.environ.get(CIRRO_AUTH_CLIENT_ID)

    os.environ[CIRRO_SERVE] = 'true'
    os.environ[CIRRO_JOB_TYPE + 'de'] = 'cirrocumulus.job_api.run_de'
    if auth_client_id is None:
        app.config[CIRRO_AUTH] = NoAuth()
    else:
        from cirrocumulus.google_auth import GoogleAuth
        app.config[CIRRO_AUTH] = GoogleAuth()
    if os.environ.get(CIRRO_DATABASE_CLASS) is None:
        os.environ[CIRRO_DATABASE_CLASS] = 'cirrocumulus.mongo_db.MongoDb'
    if os.environ[CIRRO_DB_URI] == '':
        os.environ[CIRRO_DATABASE_CLASS] = 'cirrocumulus.local_db_api.LocalDbAPI'

    app.config[CIRRO_DATABASE] = create_instance(os.environ[CIRRO_DATABASE_CLASS])
    os.environ[CIRRO_DATASET_PROVIDERS] = ','.join(['cirrocumulus.parquet_dataset.ParquetDataset',
                                                    'cirrocumulus.zarr_dataset.ZarrDataset',
                                                    'cirrocumulus.tiledb_dataset.TileDBDataset'])
    add_dataset_providers()


def main(argsv):
    import argparse
    import os
    parser = argparse.ArgumentParser(description='Run cirrocumulus server')
    parser.add_argument('--db_uri', help='Database connection URI', default=DEFAULT_DB_URI)
    parser.add_argument('-w', '--workers', dest='workers', help='The number of worker processes', type=int)
    parser.add_argument('-t', '--timeout', dest='timeout',
                        help='Workers silent for more than this many seconds are killed and restarted', type=int,
                        default=30)
    parser.add_argument('-b', '--bind', dest='bind',
                        help='Server socket to bind. Server sockets can be any of $(HOST), $(HOST):$(PORT), fd://$(FD), or unix:$(PATH). An IP is a valid $(HOST).',
                        default='127.0.0.1:5000')
    parser.add_argument('--footer', help='Markdown file to customize the application footer')
    parser.add_argument('--header', help='Markdown file to customize the application header')
    parser.add_argument('--upload', help='URL to allow users to upload files')
    parser.add_argument('--results', help='URL to save user computed results (e.g. differential expression) to')
    parser.add_argument('--ontology', help='Path to ontology in OBO format for annotation')
    args = parser.parse_args(argsv)

    bind = args.bind if args.bind is not None else '127.0.0.1:5000'
    if args.ontology is not None:
        os.environ[CIRRO_CELL_ONTOLOGY] = args.ontology
    os.environ[CIRRO_DB_URI] = args.db_uri

    if args.footer is not None:
        os.environ[CIRRO_FOOTER] = args.footer

    if args.header is not None:
        os.environ[CIRRO_BRAND] = args.header

    if args.workers is not None:
        workers = args.workers
    else:
        import os
        workers = 2 * os.cpu_count()
    if args.upload is not None:
        os.environ[CIRRO_UPLOAD] = args.upload
    if args.results is not None:
        os.environ[CIRRO_JOB_RESULTS] = args.results
        get_fs(os.environ[CIRRO_JOB_RESULTS]).makedirs(os.environ[CIRRO_JOB_RESULTS], exist_ok=True)

    run_args = [
        'gunicorn',
        '-b', bind,
        '-w', str(workers),
        '-t', str(args.timeout),
        '-n', 'cirrocumulus-webserver',
        'cirrocumulus.serve:cached_app()'
    ]
    # if args.gunicorn is not None:
    #     run_args += args.gunicorn.split(' ')
    import subprocess
    subprocess.check_call(run_args)


if __name__ == "__main__":
    main()



import datetime
import json
import os

from cirrocumulus.abstract_db import AbstractDB
from cirrocumulus.envir import CIRRO_JOB_RESULTS, SERVER_CAPABILITY_EDIT_DATASET, SERVER_CAPABILITY_ADD_DATASET, \
    SERVER_CAPABILITY_DELETE_DATASET
from cirrocumulus.io_util import unique_id
from cirrocumulus.job_api import save_job_result_to_file
from cirrocumulus.util import get_fs


def create_dataset_meta(path):
    result = {'id': path, 'url': path, 'name': os.path.splitext(os.path.basename(path.rstrip('/')))[0],
              'description': ''}
    if os.path.basename(path).endswith('.json'):
        with open(path, 'rt') as f:
            result.update(json.load(f))

    return result


def write_json(json_data, json_path):
    if os.path.exists(os.path.dirname(os.path.abspath(json_path))):  # only support writing local files
        with open(json_path, 'wt') as f:
            json.dump(json_data, f)
    else:
        print('Skipping {}'.format(json_path))


class LocalDbAPI(AbstractDB):

    def __init__(self, paths):
        super().__init__()
        self.dataset_to_info = {}  # json_data, meta, json_path
        self.job_id_to_job = {}
        if os.environ.get(CIRRO_JOB_RESULTS) is not None:  # load saved on disk
            fs = get_fs(os.environ[CIRRO_JOB_RESULTS])
            fs.makedirs(os.environ[CIRRO_JOB_RESULTS], exist_ok=True)
            for url in fs.ls(os.environ[CIRRO_JOB_RESULTS]):
                if url.lower().endswith('.json.gz'):
                    import gzip
                    with gzip.open(fs.open(url)) as f:
                        d = json.load(f)
                        if 'id' in d:
                            d['url'] = url
                            self.job_id_to_job[d['id']] = d
                elif url.lower().endswith('.json'):
                    with fs.open(url) as f:
                        d = json.load(f)
                        if 'id' in d:
                            d['url'] = url
                            self.job_id_to_job[d['id']] = d

        for path in paths:
            json_data = {}
            basename = os.path.splitext(path)[0]
            old_path = basename + '_filters.json'
            json_path = basename + '.json'
            if os.path.exists(old_path) and os.path.getsize(old_path) > 0:
                with open(old_path, 'rt') as f:
                    json_data['filters'] = json.load(f)

            if os.path.exists(json_path) and os.path.getsize(json_path) > 0:
                with open(json_path, 'rt') as f:
                    try:
                        json_data.update(json.load(f))
                    except:
                        print('Unable to load {}'.format(json_path))
            meta = create_dataset_meta(path)
            if 'filters' not in json_data:
                json_data['filters'] = {}
            if 'views' not in json_data:
                json_data['views'] = {}
            if 'categories' not in json_data:
                json_data['categories'] = {}
            self.dataset_to_info[path] = dict(json_data=json_data, meta=meta, json_path=json_path)

    def capabilities(self):
        c = super().capabilities()
        c[SERVER_CAPABILITY_EDIT_DATASET] = False
        c[SERVER_CAPABILITY_ADD_DATASET] = False
        c[SERVER_CAPABILITY_DELETE_DATASET] = False
        return c

    def __delete_entity(self, dataset_id, entity_id, kind):
        json_data = self.dataset_to_info[dataset_id]['json_data']
        del json_data[kind][entity_id]
        write_json(json_data, self.dataset_to_info[dataset_id]['json_path'])

    def __get_entity(self, dataset_id, entity_id, kind):
        json_data = self.dataset_to_info[dataset_id]['json_data']
        return json_data[kind][entity_id]

    def __get_entity_list(self, dataset_id, kind):
        info = self.dataset_to_info.get(dataset_id)
        if info is None:
            return []
        json_data = info['json_data']
        results = []
        filters = json_data[kind]
        for key in filters:
            r = filters[key]
            r['id'] = key
            results.append(r)
        return results

    def __find_dataset_id(self, entity_id, kind):
        for dataset_id in self.dataset_to_info:
            info = self.dataset_to_info.get(dataset_id)
            if info is None:
                continue
            json_data = info['json_data']
            filters = json_data[kind]
            for key in filters:
                if key == entity_id:
                    return dataset_id
        raise ValueError('{} not found'.format(entity_id))

    def __upsert_entity(self, dataset_id, entity_id, kind, entity_dict):
        if entity_id is None:
            entity_id = unique_id()
        json_data = self.dataset_to_info[dataset_id]['json_data']
        entity = json_data[kind].get(entity_id)
        if entity is None:
            entity = {}
            json_data[kind][entity_id] = entity
        entity.update(entity_dict)
        write_json(json_data, self.dataset_to_info[dataset_id]['json_path'])
        return entity_id

    def user(self, email):
        return dict()

    def datasets(self, email):
        results = []
        for key in self.dataset_to_info:
            m = self.dataset_to_info[key]['meta']
            result = {'id': m['id'], 'url': m['url'], 'name': m['name'], 'description': m.get('description')}
            results.append(result)
        return results

    def get_dataset(self, email, dataset_id, ensure_owner=False):
        info = self.dataset_to_info.get(dataset_id)
        if info is None:  # on the fly
            result = {'id': dataset_id, 'url': dataset_id, 'name': os.path.splitext(os.path.basename(dataset_id))[0],
                      'description': ''}
        else:
            result = info['meta']
            result['id'] = dataset_id

        return result

    def category_names(self, email, dataset_id):
        info = self.dataset_to_info.get(dataset_id)
        if info is None:
            return {}
        json_data = info['json_data']
        categories = json_data['categories']
        results = {}
        for category_name in categories:
            category = categories[category_name]
            cat_results = {}
            results[category_name] = cat_results
            for original_value in category:
                val = category[original_value]
                cat_results[original_value] = dict(originalValue=original_value, newValue=val.get('new'),
                                                   positiveMarkers=val.get('positiveMarkers'),
                                                   negativeMarkers=val.get('negativeMarkers'),
                                                   color=val.get('color'))
        return results

    def upsert_category_name(self, email, dataset_id, category, original_value, update):
        json_data = self.dataset_to_info[dataset_id]['json_data']
        category_entity = json_data['categories'].get(category)
        if category_entity is None:
            category_entity = {}
            json_data['categories'][category] = category_entity
        if 'newValue' in update:
            update['new'] = update.pop('newValue')
        entity = category_entity.get(original_value)
        if entity is None:
            entity = {}
        entity.update(update)
        category_entity[original_value] = entity
        if dataset_id is not None:
            entity['dataset_id'] = dataset_id
        if email is not None:
            entity['email'] = email
        write_json(json_data, self.dataset_to_info[dataset_id]['json_path'])

    def get_feature_sets(self, email, dataset_id):
        info = self.dataset_to_info.get(dataset_id)
        if info is None:
            return []
        json_data = info['json_data']
        return json_data.get('markers', [])

    def delete_feature_set(self, email, dataset_id, set_id):
        json_data = self.dataset_to_info[dataset_id]['json_data']
        markers = json_data['markers']
        for i in range(len(markers)):
            if markers[i]['id'] == set_id:
                markers.pop(i)
                break
        write_json(json_data, self.dataset_to_info[dataset_id]['json_path'])

    def upsert_feature_set(self, email, dataset_id, set_id, category, name, features):
        if set_id is None:
            set_id = unique_id()
        else:
            self.delete_feature_set(email=email, dataset_id=dataset_id, set_id=set_id)
        json_data = self.dataset_to_info[dataset_id]['json_data']
        markers = json_data.get('markers')
        if markers is None:
            markers = []
            json_data['markers'] = markers
        markers.append(dict(id=set_id, features=features, name=name, category=category))
        write_json(json_data, self.dataset_to_info[dataset_id]['json_path'])
        return set_id

    def dataset_views(self, email, dataset_id):
        return self.__get_entity_list(dataset_id=dataset_id, kind='views')

    def delete_dataset_view(self, email, view_id):
        dataset_id = self.__find_dataset_id(view_id, 'views')
        return self.__delete_entity(dataset_id=dataset_id, entity_id=view_id, kind='views')

    def get_dataset_view(self, email, view_id):
        dataset_id = self.__find_dataset_id(view_id, 'views')
        return self.__get_entity(dataset_id=dataset_id, entity_id=view_id, kind='views')

    def upsert_dataset_view(self, email, dataset_id, view):
        view['last_updated'] = str(datetime.datetime.utcnow())
        if 'value' in view:
            view['value'] = json.dumps(view['value'])
        if email is not None:
            view['email'] = email
        if dataset_id is not None:
            view['dataset_id'] = dataset_id
        view_id = view.pop('id') if 'id' in view else None
        view_id = self.__upsert_entity(dataset_id=dataset_id, entity_id=view_id, kind='views', entity_dict=view)
        return dict(id=view_id, last_updated=view['last_updated'])

    def create_job(self, email, dataset_id, job_name, job_type, params):
        job_id = unique_id()
        self.job_id_to_job[job_id] = dict(id=job_id, dataset_id=dataset_id, name=job_name, type=job_type, params=params,
                                          status=None, submitted=str(datetime.datetime.utcnow()))
        return job_id

    def get_job(self, email, job_id, return_type):
        job = self.job_id_to_job.get(job_id)
        if job is None:
            return None
        if return_type == 'result':
            return job['result'] if 'result' in job else job
        elif return_type == 'status':
            return dict(status=job['status'])
        elif return_type == 'params':
            return dict(params=job.get('params'))

    def get_jobs(self, email, dataset_id):
        results = []
        for job in self.job_id_to_job.values():
            if dataset_id == job['dataset_id']:
                results.append(dict(id=job['id'], name=job['name'], type=job['type'], status=job['status'],
                                    submitted=job['submitted']))
        return results

    def delete_job(self, email, job_id):
        job = self.job_id_to_job.pop(job_id)
        if 'url' in job and os.path.exists(job['url']):
            os.remove(job['url'])

    def update_job(self, email, job_id, status, result):
        job = self.job_id_to_job[job_id]
        job['status'] = status
        if result is not None:
            if os.environ.get(CIRRO_JOB_RESULTS) is not None:  # save to directory
                result.update(job)
                save_job_result_to_file(result, job_id)
                job['result'] = result
            else:
                from cirrocumulus.util import to_json
                job['result'] = to_json(result)



import json

import zarr

from cirrocumulus.abstract_backed_dataset import AbstractBackedDataset
from cirrocumulus.anndata_util import dataset_schema


class ZarrDataset(AbstractBackedDataset):

    def __init__(self):
        super().__init__()

    def get_suffixes(self):
        return ['zarr']

    def is_group(self, node):
        return isinstance(node, zarr.hierarchy.Group)

    def open_group(self, filesystem, path):
        return zarr.open_group(filesystem.get_mapper(path), mode='r')

    def slice_dense_array(self, X, indices):
        return X.get_orthogonal_selection((slice(None), indices))

    def get_schema(self, filesystem, path):
        g = zarr.open_group(filesystem.get_mapper(path), mode='r')
        return json.loads(str(g['uns']['cirro-schema'][...])) if 'cirro-schema' in g['uns'] else dataset_schema(g,
                                                                                                                n_features=0)
